import click

from .downgrade import downgrade_
from .init import init_
from .revision import revision_
from .upgrade import upgrade_


@click.group()
def migration():
    """Database migrations"""

    pass


migration.add_command(downgrade_)
migration.add_command(init_)
migration.add_command(revision_)
migration.add_command(upgrade_)



import click

from .create_user import create_user
from .update_user import update_user


@click.group()
def user():
    """Manage users"""
    pass


user.add_command(create_user)
user.add_command(update_user)



from datetime import datetime, timedelta

import jwt

from mvapi.settings import settings
from mvapi.web.libs.exceptions import JWTError
from mvapi.web.libs.misc import JSONEncoder
from mvapi.web.models.session import Session


class JSONWebToken:
    token_type = 'Bearer'
    expires = None

    __algorithm = None
    __secret_key = None
    __expires_from = None

    def __init__(self, expires_from=None, temp=False):
        jwt_settings = settings.JWTAUTH_SETTINGS

        self.__algorithm = jwt_settings.get('ALGORITHM', 'HS256')
        self.__secret_key = settings.SECRET_KEY
        self.__expires_from = expires_from or datetime.utcnow()

        if temp:
            self.expires = self.__get_expires(hours=1)
        else:
            days = jwt_settings.get('EXPIRES', 365)
            self.expires = self.__get_expires(days=days)

    def __decode_token(self, token):
        return jwt.decode(
            token, self.__secret_key, algorithms=[self.__algorithm]
        )

    def __get_expires(self, days=0, hours=0):
        return self.__expires_from + timedelta(days=days, hours=hours)

    def get_token(self, session: Session):
        payload = {
            'session_id': session.id_,
            'exp': self.expires
        }

        return jwt.encode(payload, key=self.__secret_key,
                          algorithm=self.__algorithm,
                          json_encoder=JSONEncoder)

    def get_user(self, token):
        try:
            payload = self.__decode_token(token)

        except jwt.ExpiredSignatureError:
            raise JWTError('Token has expired')

        except jwt.DecodeError:
            raise JWTError('Error decoding token')

        session_id = payload.get('session_id')
        if not session_id:
            raise JWTError('Invalid payload')

        session = Session.query.get(session_id)
        return session.user



import logging


def init_logger(verbosity):
    logging.basicConfig(
        level=logging.DEBUG if verbosity else logging.INFO,
        format='%(asctime)s %(name)s %(levelname)s: %(message)s',
    )



import click
from alembic.command import revision

from .utils import DIRECTORY, get_config


@click.command('revision')
@click.option('--directory', default=DIRECTORY,
              help='path to the migrations directory')
@click.option('--message', help='message to apply to the revision')
@click.option('--autogenerate', is_flag=True, default=True,
              help='autogenerate the script from the database')
@click.option('--sql', is_flag=True,
              help='dump the script out as a SQL string; when specified, '
                   'the script is dumped to stdout')
@click.option('--head', default='head',
              help='head revision to build the new revision upon as a parent')
@click.option('--splice', is_flag=True,
              help='the new revision should be made into a new head of its '
                   'own; is required when the given `head` is not itself a '
                   'head')
@click.option('--branch-label', help='label to apply to the branch')
@click.option('--version-path', help='symbol identifying a specific version '
                                     'path from the configuration')
@click.option('--rev_id', help='revision identifier to use instead of having '
                               'one generated')
def revision_(directory, message, autogenerate, sql, head, splice, branch_label,
              version_path, rev_id):
    """Create a new revision file"""

    config = get_config(directory)
    revision(config, message=message, autogenerate=autogenerate, sql=sql,
             head=head, splice=splice, branch_label=branch_label,
             version_path=version_path, rev_id=rev_id)



import os

from alembic.config import Config as AlembicConfig

DIRECTORY = 'migrations'


class Config(AlembicConfig):
    def get_template_directory(self):
        package_dir = os.path.abspath(os.path.dirname(__file__))
        return os.path.join(package_dir, 'templates')


def get_config(directory):
    config = Config()
    config.config_file_name = os.path.join(directory, 'alembic.ini')

    return config



import click

from mvapi.libs.logger import init_logger
from mvapi.version import version
from .init_project import init_project


@click.group()
@click.version_option(version)
@click.option('--verbosity', is_flag=True, default=False, help='Show debug log')
def cli(verbosity):
    init_logger(verbosity)


cli.add_command(init_project)



import os

from dotenv import load_dotenv

from mvapi.libs.exceptions import NoSettingsModuleSpecified
from mvapi.libs.misc import import_object


def get_settings():
    config_file = os.environ.get('MVAPI_CONFIG', False)

    if config_file:
        if config_file.startswith('~'):
            config_file = os.path.expanduser(config_file)
        config_file = os.path.abspath(config_file)

        if os.path.exists(config_file):
            load_dotenv(config_file)

    app_settings = os.environ.get('SETTINGS')
    if not app_settings:
        raise NoSettingsModuleSpecified(
            'Path to settings module is not found'
        )

    settings_obj = import_object(app_settings)
    settings_obj.APP_NAME = app_settings.partition('.')[0]

    return settings_obj()


settings = get_settings()



import click

from mvapi.libs.logger import init_logger
from mvapi.libs.misc import import_object
from mvapi.settings import settings
from .migration import migration
from .run_temp_script import run_temp_script
from .user import user
from .web import web

version = import_object(f'{settings.APP_NAME}.version.version')


@click.group()
@click.version_option(version)
@click.option('--verbosity', is_flag=True, default=settings.DEBUG,
              help='Show debug log')
def cli(verbosity):
    init_logger(verbosity)


cli.add_command(migration)
cli.add_command(run_temp_script)
cli.add_command(user)
cli.add_command(web)



import importlib

import click
from dateutil.parser import parse
from jinja2 import Environment, PackageLoader, PrefixLoader


def import_object(path):
    module_name, object_name = path.rsplit('.', 1)
    mod = importlib.import_module(module_name)
    if not hasattr(mod, object_name):
        raise ImportError

    return getattr(mod, object_name)


def render_template(template_name, data):
    from mvapi.settings import settings

    prefixes = {'mvapi': PackageLoader('mvapi', 'templates')}
    for k, v in settings.TEMPLATE_LOADERS.items():
        prefixes[k] = PackageLoader(*v)

    env = Environment(loader=PrefixLoader(prefixes))
    tmpl = env.get_template(template_name)
    return tmpl.render(data)


# noinspection PyPep8Naming
class classproperty(object):
    def __init__(self, fget):
        self.fget = fget

    def __get__(self, owner_self, owner_cls):
        return self.fget(owner_cls)


# noinspection PyUnusedLocal
def validate_date_arg(ctx, param, value):
    if not value:
        return None

    try:
        return parse(value)
    except ValueError:
        raise click.BadParameter('value cannot be parsed')



import click
from alembic.command import upgrade

from .utils import DIRECTORY, get_config


@click.command('upgrade')
@click.option('--directory', default=DIRECTORY,
              help='path to the migrations directory')
@click.option('--revision', default='head',
              help='revision target or range for --sql mode')
@click.option('--sql', is_flag=True,
              help='dump the script out as a SQL string; when specified, '
                   'the script is dumped to stdout')
@click.option('--tag', help='an arbitrary tag that can be intercepted by '
                            'custom `env.py` scripts')
def upgrade_(directory, revision, sql, tag):
    """Upgrade to a later version"""

    config = get_config(directory)
    upgrade(config, revision, sql=sql, tag=tag)



import logging

import click

from mvapi.libs.exceptions import NotFoundError
from mvapi.web.models.user import User

logger = logging.getLogger(__name__)


@click.command('update-user')
@click.option('--email', required=True, help='user email')
@click.option('--is-admin', is_flag=True, help='make user an admin')
@click.password_option(help='user password')
def update_user(email, password, is_admin):
    """Update a user"""

    try:
        user = User.query.get_by(email=email)
    except NotFoundError:
        logger.info(f'User with email {email} not found')
        return

    if password:
        user.password = password

    if is_admin:
        user.is_admin = is_admin

    logger.info(f'User {user} updated')



import os


class DefaultSettings:
    BLUEPRINTS = []
    CONVERTERS = []
    DEBUG = False
    DEBUG_SQL = False
    EMAILS_MODULE = None
    ENV = 'production'
    ERRORS_PATH = '.errors'
    EXTENSIONS = []
    JWTAUTH_SETTINGS = {}
    LIMIT = 15
    MIGRATIONS_EXCLUDE_TABLES = tuple()
    MODELS = []
    SERIALIZERS = []
    SQLALCHEMY_TRACK_MODIFICATIONS = False
    TEMPLATE_LOADERS = {}
    VIEWS = []

    # noinspection PyPep8Naming
    @property
    def SQLALCHEMY_DATABASE_URI(self):
        return self.DB_URI

    def __getattr__(self, item):
        if item in os.environ:
            return os.environ[item]
        raise KeyError



class AppException(Exception):
    pass


class NoCommandException(Exception):
    pass


class NoSettingsModuleSpecified(Exception):
    pass


###

class ModelKeyError(AppException):
    pass


class NotFoundError(AppException):
    pass



from mvapi.web.models.user import User
from mvapi.web.serializers.base import BaseSerializer


class UserSerializer(BaseSerializer):
    resource_type = 'user'
    item: User = None

    def render_attributes(self):
        attrs = super(UserSerializer, self).render_attributes()

        if self.current_user:
            if self.current_user.is_admin or \
                    self.item.id_ == self.current_user.id_:
                attrs['email'] = self.item.email

            if self.current_user.is_admin:
                attrs['is_admin'] = self.item.is_admin

        return attrs



import logging

import click

from mvapi.libs.exceptions import NotFoundError
from mvapi.web.models.user import User

logger = logging.getLogger(__name__)


@click.command('create-user')
@click.option('--email', required=True, help='user email')
@click.option('--is-admin', is_flag=True, help='make user an admin')
@click.password_option(help='user password', required=True)
def create_user(email, password, is_admin):
    """Create a user"""

    try:
        user = User.query.get_by(email=email)
        logger.info(f'User {user} exists')
        return
    except NotFoundError:
        pass

    user = User.create(email=email, password=password, is_admin=is_admin)
    logger.info(f'User {user} created')



from flask import request

from mvapi.libs.exceptions import NotFoundError
from mvapi.web.libs.decorators import auth_required
from mvapi.web.libs.exceptions import BadRequestError, UnauthorizedError
from mvapi.web.models.session import Session
from mvapi.web.models.user import User
from mvapi.web.views.base import BaseView


class SessionsView(BaseView):
    resource_type = 'sessions'
    resource_model = Session

    @auth_required
    def get(self):
        return (self.current_user.sessions.get(self.resource_id)
                if self.resource_id
                else self.current_user.sessions.apply_args(**self.common_args))

    def post(self):
        data = self.available_data(required={'email'}, extra={'password'})
        password = data.get('password')

        if not password and not (self.current_user and
                                 self.current_user.is_admin):
            raise BadRequestError('Password is required')

        try:
            user = User.query.get_by(email=data['email'])
        except NotFoundError:
            raise UnauthorizedError('Email address not found')

        if password and not user.passwords_matched(password=password):
            raise UnauthorizedError('Password is wrong')

        return Session.create(
            user=user,
            remote_ip=request.remote_addr,
            user_agent=request.user_agent.string
        )
